@@ -17,6 +17,8 @@
 #include <linux/kernel.h>
 #include <linux/init.h>
 #include <linux/delay.h>
+#include <pcmcia/cs_types.h>
+#include <pcmcia/cs.h>
 #include <pcmcia/cistpl.h>
 #include <pcmcia/cisreg.h>
 #include <pcmcia/ds.h>
@@ -96,6 +98,14 @@ orinoco_cs_hard_reset(struct orinoco_private *priv)
 /* PCMCIA stuff     						    */
 /********************************************************************/
 
+/*
+ * This creates an "instance" of the driver, allocating local data
+ * structures for one device.  The device is registered with Card
+ * Services.
+ *
+ * The dev_link structure is initialized, but we don't actually
+ * configure the card at this point -- we wait until we receive a card
+ * insertion event.  */
 static int
 orinoco_cs_probe(struct pcmcia_device *link)
 {
@@ -117,14 +127,24 @@ orinoco_cs_probe(struct pcmcia_device *link)
 	link->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING;
 	link->irq.Handler = orinoco_interrupt;
 #endif
-#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37))
+
+	/* General socket configuration defaults can go here.  In this
+	 * client, we assume very little, and rely on the CIS for
+	 * almost everything.  In most clients, many details (i.e.,
+	 * number, sizes, and attributes of IO windows) are fixed by
+	 * the nature of the device, and can be hard-wired here. */
 	link->conf.Attributes = 0;
 	link->conf.IntType = INT_MEMORY_AND_IO;
-#endif
 
 	return orinoco_cs_config(link);
 }				/* orinoco_cs_attach */
 
+/*
+ * This deletes a driver "instance".  The device is de-registered with
+ * Card Services.  If it has been released, all local data structures
+ * are freed.  Otherwise, the structures will be freed when the device
+ * is released.
+ */
 static void orinoco_cs_detach(struct pcmcia_device *link)
 {
 	struct orinoco_private *priv = link->priv;
@@ -136,15 +156,12 @@ static void orinoco_cs_detach(struct pcmcia_device *link)
 	free_orinocodev(priv);
 }				/* orinoco_cs_detach */
 
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37))
-static int orinoco_cs_config_check(struct pcmcia_device *p_dev, void *priv_data)
-{
-	if (p_dev->config_index == 0)
-		return -EINVAL;
+/*
+ * orinoco_cs_config() is scheduled to run after a CARD_INSERTION
+ * event is received, to configure the PCMCIA socket, and to make the
+ * device available to the system.
+ */
 
-	return pcmcia_request_io(p_dev);
-};
-#else
 static int orinoco_cs_config_check(struct pcmcia_device *p_dev,
 				   cistpl_cftable_entry_t *cfg,
 				   cistpl_cftable_entry_t *dflt,
@@ -185,21 +202,9 @@ static int orinoco_cs_config_check(struct pcmcia_device *p_dev,
 	p_dev->conf.Attributes |= CONF_ENABLE_IRQ;
 
 	/* IO window settings */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-	p_dev->resource[0]->end = p_dev->resource[1]->end = 0;
-#else
 	p_dev->io.NumPorts1 = p_dev->io.NumPorts2 = 0;
-#endif
 	if ((cfg->io.nwin > 0) || (dflt->io.nwin > 0)) {
 		cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt->io;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-		p_dev->io_lines = io->flags & CISTPL_IO_LINES_MASK;
-		p_dev->resource[0]->flags &= ~IO_DATA_PATH_WIDTH;
-		p_dev->resource[0]->flags |=
-			pcmcia_io_cfg_data_width(io->flags);
-		p_dev->resource[0]->start = io->win[0].base;
-		p_dev->resource[0]->end = io->win[0].len;
-#else
 		p_dev->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
 		if (!(io->flags & CISTPL_IO_8BIT))
 			p_dev->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
@@ -208,25 +213,14 @@ static int orinoco_cs_config_check(struct pcmcia_device *p_dev,
 		p_dev->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
 		p_dev->io.BasePort1 = io->win[0].base;
 		p_dev->io.NumPorts1 = io->win[0].len;
-#endif
 		if (io->nwin > 1) {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-			p_dev->resource[1]->flags = p_dev->resource[0]->flags;
-			p_dev->resource[1]->start = io->win[1].base;
-			p_dev->resource[1]->end = io->win[1].len;
-#else
 			p_dev->io.Attributes2 = p_dev->io.Attributes1;
 			p_dev->io.BasePort2 = io->win[1].base;
 			p_dev->io.NumPorts2 = io->win[1].len;
-#endif
 		}
 
 		/* This reserves IO space but doesn't actually enable it */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-		if (pcmcia_request_io(p_dev) != 0)
-#else
 		if (pcmcia_request_io(p_dev, &p_dev->io) != 0)
-#endif
 			goto next_entry;
 	}
 	return 0;
@@ -235,7 +229,6 @@ next_entry:
 	pcmcia_disable_device(p_dev);
 	return -ENODEV;
 };
-#endif
 
 static int
 orinoco_cs_config(struct pcmcia_device *link)
@@ -245,12 +238,20 @@ orinoco_cs_config(struct pcmcia_device *link)
 	int ret;
 	void __iomem *mem;
 
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37))
-	link->config_flags |= CONF_AUTO_SET_VPP | CONF_AUTO_CHECK_VCC |
-		CONF_AUTO_SET_IO | CONF_ENABLE_IRQ;
-	if (ignore_cis_vcc)
-		link->config_flags &= ~CONF_AUTO_CHECK_VCC;
-#endif
+	/*
+	 * In this loop, we scan the CIS for configuration table
+	 * entries, each of which describes a valid card
+	 * configuration, including voltage, IO window, memory window,
+	 * and interrupt settings.
+	 *
+	 * We make no assumptions about the card to be configured: we
+	 * use just the information available in the CIS.  In an ideal
+	 * world, this would work for any PCMCIA card, but it requires
+	 * a complete and accurate CIS.  In practice, a driver usually
+	 * "knows" most of these things without consulting the CIS,
+	 * and most client drivers will only use the CIS to fill in
+	 * implementation-defined details.
+	 */
 	ret = pcmcia_loop_config(link, orinoco_cs_config_check, NULL);
 	if (ret) {
 		if (!ignore_cis_vcc)
@@ -271,18 +272,18 @@ orinoco_cs_config(struct pcmcia_device *link)
 	/* We initialize the hermes structure before completing PCMCIA
 	 * configuration just in case the interrupt handler gets
 	 * called. */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-	mem = ioport_map(link->resource[0]->start,
-			resource_size(link->resource[0]));
-#else
 	mem = ioport_map(link->io.BasePort1, link->io.NumPorts1);
-#endif
 	if (!mem)
 		goto failed;
 
 	hermes_struct_init(hw, mem, HERMES_16BIT_REGSPACING);
 
-	ret = pcmcia_enable_device(link);
+	/*
+	 * This actually configures the PCMCIA socket -- setting up
+	 * the I/O windows and the interrupt mapping, and putting the
+	 * card and host interface into "Memory and IO" mode.
+	 */
+	ret = pcmcia_request_configuration(link, &link->conf);
 	if (ret)
 		goto failed;
 
@@ -293,14 +294,10 @@ orinoco_cs_config(struct pcmcia_device *link)
 	}
 
 	/* Register an interface with the stack */
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
-	if (orinoco_if_add(priv, link->resource[0]->start,
-			   link->irq, NULL) != 0) {
-#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35))
 	if (orinoco_if_add(priv, link->io.BasePort1,
+#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35))
 			   link->irq, NULL) != 0) {
 #else
-	if (orinoco_if_add(priv, link->io.BasePort1,
 			   link->irq.AssignedIRQ, NULL) != 0) {
 #endif
 		printk(KERN_ERR PFX "orinoco_if_add() failed\n");
@@ -314,6 +311,11 @@ orinoco_cs_config(struct pcmcia_device *link)
 	return -ENODEV;
 }				/* orinoco_cs_config */
 
+/*
+ * After a card is removed, orinoco_cs_release() will unregister the
+ * device, and release the PCMCIA configuration.  If the device is
+ * still open, this will be postponed until it is closed.
+ */
 static void
 orinoco_cs_release(struct pcmcia_device *link)
 {
@@ -362,6 +364,12 @@ static int orinoco_cs_resume(struct pcmcia_device *link)
 /* Module initialization					    */
 /********************************************************************/
 
+/* Can't be declared "const" or the whole __initdata section will
+ * become const */
+static char version[] __initdata = DRIVER_NAME " " DRIVER_VERSION
+	" (David Gibson <hermes@gibson.dropbear.id.au>, "
+	"Pavel Roskin <proski@gnu.org>, et al)";
+
 static struct pcmcia_device_id orinoco_cs_ids[] = {
 	PCMCIA_DEVICE_MANF_CARD(0x0101, 0x0777), /* 3Com AirConnect PCI 777A */
 	PCMCIA_DEVICE_MANF_CARD(0x0156, 0x0002), /* Lucent Orinoco and old Intersil */
@@ -453,13 +461,9 @@ MODULE_DEVICE_TABLE(pcmcia, orinoco_cs_ids);
 
 static struct pcmcia_driver orinoco_driver = {
 	.owner		= THIS_MODULE,
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37))
-	.name		= DRIVER_NAME,
-#else
 	.drv		= {
 		.name	= DRIVER_NAME,
 	},
-#endif
 	.probe		= orinoco_cs_probe,
 	.remove		= orinoco_cs_detach,
 	.id_table       = orinoco_cs_ids,
@@ -470,6 +474,8 @@ static struct pcmcia_driver orinoco_driver = {
 static int __init
 init_orinoco_cs(void)
 {
+	printk(KERN_DEBUG "%s\n", version);
+
 	return pcmcia_register_driver(&orinoco_driver);
 }
 
