@@ -25,6 +25,8 @@
 #include <linux/kernel.h>
 #include <linux/init.h>
 #include <linux/delay.h>
+#include <pcmcia/cs_types.h>
+#include <pcmcia/cs.h>
 #include <pcmcia/cistpl.h>
 #include <pcmcia/cisreg.h>
 #include <pcmcia/ds.h>
@@ -78,27 +80,35 @@ static int
 spectrum_reset(struct pcmcia_device *link, int idle)
 {
 	int ret;
-	u8 save_cor;
-	u8 ccsr;
+	conf_reg_t reg;
+	u_int save_cor;
 
 	/* Doing it if hardware is gone is guaranteed crash */
 	if (!pcmcia_dev_present(link))
 		return -ENODEV;
 
 	/* Save original COR value */
-	ret = pcmcia_read_config_byte(link, CISREG_COR, &save_cor);
+	reg.Function = 0;
+	reg.Action = CS_READ;
+	reg.Offset = CISREG_COR;
+	ret = pcmcia_access_configuration_register(link, &reg);
 	if (ret)
 		goto failed;
+	save_cor = reg.Value;
 
 	/* Soft-Reset card */
-	ret = pcmcia_write_config_byte(link, CISREG_COR,
-				(save_cor | COR_SOFT_RESET));
+	reg.Action = CS_WRITE;
+	reg.Offset = CISREG_COR;
+	reg.Value = (save_cor | COR_SOFT_RESET);
+	ret = pcmcia_access_configuration_register(link, &reg);
 	if (ret)
 		goto failed;
 	udelay(1000);
 
 	/* Read CCSR */
-	ret = pcmcia_read_config_byte(link, CISREG_CCSR, &ccsr);
+	reg.Action = CS_READ;
+	reg.Offset = CISREG_CCSR;
+	ret = pcmcia_access_configuration_register(link, &reg);
 	if (ret)
 		goto failed;
 
@@ -106,15 +116,19 @@ spectrum_reset(struct pcmcia_device *link, int idle)
 	 * Start or stop the firmware.  Memory width bit should be
 	 * preserved from the value we've just read.
 	 */
-	ccsr = (idle ? HCR_IDLE : HCR_RUN) | (ccsr & HCR_MEM16);
-	ret = pcmcia_write_config_byte(link, CISREG_CCSR, ccsr);
+	reg.Action = CS_WRITE;
+	reg.Offset = CISREG_CCSR;
+	reg.Value = (idle ? HCR_IDLE : HCR_RUN) | (reg.Value & HCR_MEM16);
+	ret = pcmcia_access_configuration_register(link, &reg);
 	if (ret)
 		goto failed;
 	udelay(1000);
 
 	/* Restore original COR configuration index */
-	ret = pcmcia_write_config_byte(link, CISREG_COR,
-				(save_cor & ~COR_SOFT_RESET));
+	reg.Action = CS_WRITE;
+	reg.Offset = CISREG_COR;
+	reg.Value = (save_cor & ~COR_SOFT_RESET);
+	ret = pcmcia_access_configuration_register(link, &reg);
 	if (ret)
 		goto failed;
 	udelay(1000);
@@ -153,6 +167,14 @@ spectrum_cs_stop_firmware(struct orinoco_private *priv, int idle)
 /* PCMCIA stuff     						    */
 /********************************************************************/
 
+/*
+ * This creates an "instance" of the driver, allocating local data
+ * structures for one device.  The device is registered with Card
+ * Services.
+ *
+ * The dev_link structure is initialized, but we don't actually
+ * configure the card at this point -- we wait until we receive a card
+ * insertion event.  */
 static int
 spectrum_cs_probe(struct pcmcia_device *link)
 {
@@ -175,14 +197,24 @@ spectrum_cs_probe(struct pcmcia_device *link)
 	link->irq.Attributes = IRQ_TYPE_DYNAMIC_SHARING;
 	link->irq.Handler = orinoco_interrupt;
 #endif
-#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,37))
+
+	/* General socket configuration defaults can go here.  In this
+	 * client, we assume very little, and rely on the CIS for
+	 * almost everything.  In most clients, many details (i.e.,
+	 * number, sizes, and attributes of IO windows) are fixed by
+	 * the nature of the device, and can be hard-wired here. */
 	link->conf.Attributes = 0;
 	link->conf.IntType = INT_MEMORY_AND_IO;
-#endif
 
 	return spectrum_cs_config(link);
 }				/* spectrum_cs_attach */
 
+/*
+ * This deletes a driver "instance".  The device is de-registered with
+ * Card Services.  If it has been released, all local data structures
+ * are freed.  Otherwise, the structures will be freed when the device
+ * is released.
+ */
 static void spectrum_cs_detach(struct pcmcia_device *link)
 {
 	struct orinoco_private *priv = link->priv;
@@ -194,16 +226,12 @@ static void spectrum_cs_detach(struct pcmcia_device *link)
 	free_orinocodev(priv);
 }				/* spectrum_cs_detach */
 
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37))
-static int spectrum_cs_config_check(struct pcmcia_device *p_dev,
-				    void *priv_data)
-{
-	if (p_dev->config_index == 0)
-		return -EINVAL;
+/*
+ * spectrum_cs_config() is scheduled to run after a CARD_INSERTION
+ * event is received, to configure the PCMCIA socket, and to make the
+ * device available to the system.
+ */
 
-	return pcmcia_request_io(p_dev);
-};
-#else
 static int spectrum_cs_config_check(struct pcmcia_device *p_dev,
 				    cistpl_cftable_entry_t *cfg,
 				    cistpl_cftable_entry_t *dflt,
@@ -244,21 +272,9 @@ static int spectrum_cs_config_check(struct pcmcia_device *p_dev,
 	p_dev->conf.Attributes |= CONF_ENABLE_IRQ;
 
 	/* IO window settings */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-	p_dev->resource[0]->end = p_dev->resource[1]->end = 0;
-#else
 	p_dev->io.NumPorts1 = p_dev->io.NumPorts2 = 0;
-#endif
 	if ((cfg->io.nwin > 0) || (dflt->io.nwin > 0)) {
 		cistpl_io_t *io = (cfg->io.nwin) ? &cfg->io : &dflt->io;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-		p_dev->io_lines = io->flags & CISTPL_IO_LINES_MASK;
-		p_dev->resource[0]->flags &= ~IO_DATA_PATH_WIDTH;
-		p_dev->resource[0]->flags |=
-			pcmcia_io_cfg_data_width(io->flags);
-		p_dev->resource[0]->start = io->win[0].base;
-		p_dev->resource[0]->end = io->win[0].len;
-#else
 		p_dev->io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
 		if (!(io->flags & CISTPL_IO_8BIT))
 			p_dev->io.Attributes1 = IO_DATA_PATH_WIDTH_16;
@@ -267,25 +283,14 @@ static int spectrum_cs_config_check(struct pcmcia_device *p_dev,
 		p_dev->io.IOAddrLines = io->flags & CISTPL_IO_LINES_MASK;
 		p_dev->io.BasePort1 = io->win[0].base;
 		p_dev->io.NumPorts1 = io->win[0].len;
-#endif
 		if (io->nwin > 1) {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-			p_dev->resource[1]->flags = p_dev->resource[0]->flags;
-			p_dev->resource[1]->start = io->win[1].base;
-			p_dev->resource[1]->end = io->win[1].len;
-#else
 			p_dev->io.Attributes2 = p_dev->io.Attributes1;
 			p_dev->io.BasePort2 = io->win[1].base;
 			p_dev->io.NumPorts2 = io->win[1].len;
-#endif
 		}
 
 		/* This reserves IO space but doesn't actually enable it */
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-		if (pcmcia_request_io(p_dev) != 0)
-#else
 		if (pcmcia_request_io(p_dev, &p_dev->io) != 0)
-#endif
 			goto next_entry;
 	}
 	return 0;
@@ -294,7 +299,6 @@ next_entry:
 	pcmcia_disable_device(p_dev);
 	return -ENODEV;
 };
-#endif
 
 static int
 spectrum_cs_config(struct pcmcia_device *link)
@@ -304,12 +308,20 @@ spectrum_cs_config(struct pcmcia_device *link)
 	int ret;
 	void __iomem *mem;
 
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37))
-	link->config_flags |= CONF_AUTO_SET_VPP | CONF_AUTO_CHECK_VCC |
-		CONF_AUTO_SET_IO | CONF_ENABLE_IRQ;
-	if (ignore_cis_vcc)
-		link->config_flags &= ~CONF_AUTO_CHECK_VCC;
-#endif
+	/*
+	 * In this loop, we scan the CIS for configuration table
+	 * entries, each of which describes a valid card
+	 * configuration, including voltage, IO window, memory window,
+	 * and interrupt settings.
+	 *
+	 * We make no assumptions about the card to be configured: we
+	 * use just the information available in the CIS.  In an ideal
+	 * world, this would work for any PCMCIA card, but it requires
+	 * a complete and accurate CIS.  In practice, a driver usually
+	 * "knows" most of these things without consulting the CIS,
+	 * and most client drivers will only use the CIS to fill in
+	 * implementation-defined details.
+	 */
 	ret = pcmcia_loop_config(link, spectrum_cs_config_check, NULL);
 	if (ret) {
 		if (!ignore_cis_vcc)
@@ -319,21 +331,6 @@ spectrum_cs_config(struct pcmcia_device *link)
 		goto failed;
 	}
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
-	mem = ioport_map(link->resource[0]->start,
-			resource_size(link->resource[0]));
-#else
-	mem = ioport_map(link->io.BasePort1, link->io.NumPorts1);
-#endif
-	if (!mem)
-		goto failed;
-
-	/* We initialize the hermes structure before completing PCMCIA
-	 * configuration just in case the interrupt handler gets
-	 * called. */
-	hermes_struct_init(hw, mem, HERMES_16BIT_REGSPACING);
-	hw->eeprom_pda = true;
-
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35))
 	ret = pcmcia_request_irq(link, orinoco_interrupt);
 #else
@@ -342,7 +339,22 @@ spectrum_cs_config(struct pcmcia_device *link)
 	if (ret)
 		goto failed;
 
-	ret = pcmcia_enable_device(link);
+	/* We initialize the hermes structure before completing PCMCIA
+	 * configuration just in case the interrupt handler gets
+	 * called. */
+	mem = ioport_map(link->io.BasePort1, link->io.NumPorts1);
+	if (!mem)
+		goto failed;
+
+	hermes_struct_init(hw, mem, HERMES_16BIT_REGSPACING);
+	hw->eeprom_pda = true;
+
+	/*
+	 * This actually configures the PCMCIA socket -- setting up
+	 * the I/O windows and the interrupt mapping, and putting the
+	 * card and host interface into "Memory and IO" mode.
+	 */
+	ret = pcmcia_request_configuration(link, &link->conf);
 	if (ret)
 		goto failed;
 
@@ -357,14 +369,10 @@ spectrum_cs_config(struct pcmcia_device *link)
 	}
 
 	/* Register an interface with the stack */
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
-	if (orinoco_if_add(priv, link->resource[0]->start,
-			   link->irq, NULL) != 0) {
-#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35))
 	if (orinoco_if_add(priv, link->io.BasePort1,
+#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,35))
 			   link->irq, NULL) != 0) {
 #else
-	if (orinoco_if_add(priv, link->io.BasePort1,
 			   link->irq.AssignedIRQ, NULL) != 0) {
 #endif
 		printk(KERN_ERR PFX "orinoco_if_add() failed\n");
@@ -378,6 +386,11 @@ spectrum_cs_config(struct pcmcia_device *link)
 	return -ENODEV;
 }				/* spectrum_cs_config */
 
+/*
+ * After a card is removed, spectrum_cs_release() will unregister the
+ * device, and release the PCMCIA configuration.  If the device is
+ * still open, this will be postponed until it is closed.
+ */
 static void
 spectrum_cs_release(struct pcmcia_device *link)
 {
@@ -422,6 +435,12 @@ spectrum_cs_resume(struct pcmcia_device *link)
 /* Module initialization					    */
 /********************************************************************/
 
+/* Can't be declared "const" or the whole __initdata section will
+ * become const */
+static char version[] __initdata = DRIVER_NAME " " DRIVER_VERSION
+	" (Pavel Roskin <proski@gnu.org>,"
+	" David Gibson <hermes@gibson.dropbear.id.au>, et al)";
+
 static struct pcmcia_device_id spectrum_cs_ids[] = {
 	PCMCIA_DEVICE_MANF_CARD(0x026c, 0x0001), /* Symbol Spectrum24 LA4137 */
 	PCMCIA_DEVICE_MANF_CARD(0x0104, 0x0001), /* Socket Communications CF */
@@ -432,13 +451,9 @@ MODULE_DEVICE_TABLE(pcmcia, spectrum_cs_ids);
 
 static struct pcmcia_driver orinoco_driver = {
 	.owner		= THIS_MODULE,
-#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,37))
-	.name		= DRIVER_NAME,
-#else
 	.drv		= {
 		.name	= DRIVER_NAME,
 	},
-#endif
 	.probe		= spectrum_cs_probe,
 	.remove		= spectrum_cs_detach,
 	.suspend	= spectrum_cs_suspend,
@@ -449,6 +464,8 @@ static struct pcmcia_driver orinoco_driver = {
 static int __init
 init_spectrum_cs(void)
 {
+	printk(KERN_DEBUG "%s\n", version);
+
 	return pcmcia_register_driver(&orinoco_driver);
 }
 
